/* Madalina Ciucu 332CA */

#include <avr/io.h>
#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>

/* motors */
#define MOT_PWM_FREQ 256
#define INPUT1 PD0
#define INPUT2 PD1
#define INPUT3 PD2
#define INPUT4 PD3

#define DIST_THRESHOLD 20

void init_timer1(void) {
	/* prescaler 256, channel B, Fast PWM (non-inverting) */
	TCCR1A |= (1 << WGM11) | (1 << COM1B1) | (1 << COM1A1);
	TCCR1B |= (1 << WGM12) | (1 << WGM13) | (1 << CS11);
	
	ICR1 = MOT_PWM_FREQ;
}

void init_motors(void) {
	/* init motor1 */
	DDRD |= (1 << INPUT1) | (1 << INPUT2);
	/* init motor2 */
	DDRD |= (1 << INPUT3) | (1 << INPUT4);

	init_timer1();
}

enum action_t {
	FORWARD,
	BACK,
	STOP,
	RIGHT
};

#define for_action(act) if(action == act)
void drive_motors(float speed, enum action_t action) {
#define SPEED 0.5

	const int low = 65;
	const int high = ICR1;
	const int pwm =  low + (high - low) * speed;

	PORTD = 0;

	for_action(RIGHT) {
		PORTD &= ~(1 << INPUT1);
		PORTD |= (1 << INPUT2);
		PORTD &= ~(1 << INPUT3);
		PORTD |= (1 << INPUT4);

		OCR1A = OCR1B = pwm;
	}

	for_action(STOP) {
		PORTD |= ~(1 << INPUT1);
		PORTD |= ~(1 << INPUT2);
		PORTD |= ~(1 << INPUT3);
		PORTD |= ~(1 << INPUT4);

		OCR1A = OCR1B = 0;
	}

	for_action(FORWARD) {
		PORTD &= ~(1 << INPUT1);
		PORTD |= (1 << INPUT2);
		PORTD |= (1 << INPUT3);
		PORTD &= ~(1 << INPUT4);

		OCR1A = OCR1B = pwm;
	}

	for_action(BACK) {
		PORTD |= (1 << INPUT1);
		PORTD &= ~(1 << INPUT2);
		PORTD &= ~(1 << INPUT3);
		PORTD |= (1 << INPUT4);

		OCR1A = OCR1B = pwm;
	}
}

#define US_PORT PORTB
#define US_PIN PINB
#define US_DDR  DDRB
#define US_TRIG PB1
#define US_ECHO PB0
 
 
//Calculating width of high pulse in micro second.
#define US_Time_out  -12
#define US_NO_OBSTACLE -13

void US_init()
{
    US_DDR |= (1 << US_TRIG);
	US_DDR = ~(1 << US_ECHO);
}
 
void US_trig()
{
	US_PORT &= ~(1 << US_TRIG); //low
	_delay_us(5);
     
    US_PORT |= (1 << US_TRIG); //high
     
    _delay_us(10);
     
    US_PORT &= ~(1 << US_TRIG); //low
 
}
 
uint16_t pulse_width()
{
 
    uint32_t i, result;      // Value goes from 0 to 65536.
 
    //Wait for the rising edge
	
    for(i = 0; i < 600000; i++)
    {
        if(!(US_PIN & (1 << US_ECHO)))
			continue;      //Line is still low, so wait
        else
			break;       //High edge detected, so break.
    }
 
    if(i == 600000)
		return US_Time_out;     // No Reply found from US sensor, time lapsed
     
    //High Edge Found
 
 
    //Setup Timer1
    TCCR1A = 0X00;
    TCCR1B = (1 << CS11);      //Prescaler = Fcpu/8
    TCNT1 = 0x00;        //Again initialising the counter
 
    //Counting the width of pulse.
     
    //Now wait for the falling edge
    for(i = 0; i < 600000; i++)
    {
        if(US_PIN & (1 << US_ECHO))
        {
            if(TCNT1 > 60000) 
				break;
			else 
				continue;
        }
        else
			break;
    }
 
    if(i == 600000)
		return US_NO_OBSTACLE; //Indicates time out
 
    //Falling edge found
 
    result = TCNT1;
 
    //Stop Timer
    TCCR1B = 0x00;
 
    if(result > 60000)
		return US_NO_OBSTACLE; //No obstacle
    else
		return (result >> 1);
}

int main() {
	uint16_t r;
	init_motors();
	US_init();
	int distance;
	
	while(1) {
		//Send a trigger pulse
		US_trig();
		
		r=pulse_width(); 
        distance=(r/58);
		
		
		/* the sensor doesn't sense anything */
		if (distance < 0) {
			drive_motors(SPEED, FORWARD);
			_delay_ms(100);
		/* the distance untill collision is smaller than the THRESHOLD */ 	
		} else if (distance < DIST_THRESHOLD) {
			drive_motors(0, STOP);
			_delay_ms(200);
			drive_motors(SPEED, BACK);
			_delay_ms(200);
			drive_motors(SPEED, RIGHT);
			_delay_ms(200);
		/* the distance untill collision is bigger than the THRESHOLD */	
		} else {
			drive_motors(SPEED, FORWARD);
			_delay_ms(100);
		}
		
	}
	
	
	return 0;
}
